/*
 * Copyright (c) 2003, 2004 The University of Wroclaw.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *    3. The name of the University may not be used to endorse or promote
 *       products derived from this software without specific prior
 *       written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE UNIVERSITY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

using System.Text;
using System.Globalization;
using System;

using Nemerle.Collections;

namespace Nemerle.Compiler {

public variant Token
{
  | Tok_identifier { name : string; }
  | Tok_keyword { name : string; }
  | Tok_operator { name : string; priority : int; }

  | Tok_string_literal { value : string; }
  | Tok_char_literal { value : char; }

  | Tok_sbyte_literal { value : SByte; }
  | Tok_byte_literal { value : Byte; }
  | Tok_short_literal { value : Int16; }
  | Tok_ushort_literal { value : UInt16; }
  | Tok_integer_literal { value : int; }
  | Tok_uinteger_literal { value : UInt32; }
  | Tok_long_literal { value : Int64; }
  | Tok_ulong_literal { value : UInt64; }

  | Tok_float_literal { value : float; }
  | Tok_double_literal { value : Double; }
  | Tok_decimal_literal { value : Decimal; }
  | Tok_comment { value : string; } 

  | Tok_EOF
}

public interface ILexer
{
  get_token () : Token;
  get_location () : Location;
}

public class Lexer : ILexer
{
  file : System.IO.FileStream;
  reader : System.IO.TextReader;
  file_name : string;
  comment_store : StringBuilder;
  mutable line : int;
  mutable col : int;
  mutable putback : bool;
  mutable putbackVal : char;
  mutable isPendingChar : bool;  // is there already some first char
  mutable pendingChar : char;
  mutable negated_literal : bool;
  
  mutable last_line : int;
  mutable last_col : int;
  
  static keywords : Hashtable <string, int>;

  public class Error : System.Exception
  {
    public name : string;
    public this (name : string) { this.name <- name; }
  }


  private check_last_line_for_lf () : void
  {
    def _ = file.Seek ((-1 :> Int64), IO.SeekOrigin.End);
    
    def inp = file.ReadByte ();
    
    if (inp >= 0) {
      def ch = Convert.ToChar (inp);
      when (ch != '\n') Message.warning (Location (file_name, line, col),
                                         "no new line at the end of the file");
    } else 
      throw Error ("unexpected end of file");
      
    def _ = file.Seek ((0 :> Int64), IO.SeekOrigin.Begin);
    
    ()
  }


  public this (fn : string)
  {
    file_name <- fn;
    file <- IO.FileStream (fn, IO.FileMode.Open, IO.FileAccess.Read);
    reader <- IO.StreamReader (file, Text.Encoding.UTF8);
    when (Flags.pedantic_lexer)
      check_last_line_for_lf ();
    line <- 1;
    col <- 1;
    putback <- false;
    isPendingChar <- false;
    comment_store <- StringBuilder (300);
    negated_literal <- false;
  }

  do_read () : char
  {
    def inp = reader.Read ();
    if (inp >= 0) {
      def ch = Convert.ToChar (inp);
      match (ch) {
        | '\n' =>
          line <- line + 1;
          col <- 1
        | '\r' =>
          when (Flags.pedantic_lexer)
            Message.warning (Location (file_name, line, col),
                             "CR character found in input stream")
        | '\t' =>
          when (Flags.pedantic_lexer)
            Message.warning (Location (file_name, line, col),
                             "tab character found in input stream")
        | _ =>
          col <- col + 1
      };
      ch
    } else 
      throw Error ("unexpected end of file")
  }
        
  read () : char
  {
    if (!putback)
      do_read ()
    else {
      putback <- false;
      putbackVal
    }
  }

  peek () : char
  {
    unless (putback)
      putbackVal <- do_read ();
    putback <- true;
    putbackVal
  }

  is_id_beg (ch : char) : bool
  {
    Char.IsLetter (ch) || ch == '_'
  }

  static is_opchar (ch : char) : bool
  {
    match (ch) {
      | '=' | '<' | '>' | '@' | '^' | '|' | '&' | '+'
      | '-' | '*' | '/' | '$' | '%' | '!' | '?' | '~'
      | '.' | ':' | '#' | '[' | ']' => true
      | _ => false
    }
  }

  get_op (priority : int, first_ch : char) : Token
  {
    def buf = Text.StringBuilder (5);
    
    ignore (buf.Append (first_ch));

    while (is_opchar (peek ())) { 
      ignore (buf.Append (read ()))
    };

    def s = buf.ToString ();
    match (s) {
      | "<-" => 
        // force low priority
        Tok_operator ("<-", 9)

      | "|" | ":" | ":>" | "=>" | "=" =>
        // cannot use this as regular operator
        Tok_operator (s, 0)
        
      | _ => 
        Tok_operator (s, priority)
    }
  }

  get_number (first_ch : char) : Token
  {
    def buf = Text.StringBuilder (10);
    when (negated_literal) {
      ignore (buf.Append ('-'));
      negated_literal <- false;
    };
    mutable float_mode <- 
      if (first_ch == '.') true else false;
    mutable last_was_digit <- true;

    ignore (buf.Append (first_ch));

    // read digits and . between them if it is present
    def loop () {
      match (peek ()) {
        | '.' => 
          unless (float_mode) {
            float_mode <- true;
            ignore (read ());
            if (Char.IsDigit (peek ())) {
              ignore (buf.Append ('.'));
              loop ()
            }
            else {
              isPendingChar <- true;
              pendingChar <- '.';
              last_was_digit <- false;
              float_mode <- false;
            }
          }
        | x when Char.IsDigit (x) =>
          ignore (buf.Append (read ()));
          loop ()
        | _ => ()
      }
    };
    loop ();

    def exponent_part (only_realsuf) {
      when (!only_realsuf) {
        match (peek ()) {
          | 'E' | 'e' =>
            ignore (buf.Append (read ()));
            match (peek ()) {
              | '+' | '-' =>
                ignore (buf.Append (read ()));
              | _ => ()
            };
            if (Char.IsDigit (peek ()))
              do {
                ignore (buf.Append (read ()));
              } while (Char.IsDigit (peek ()))
            else
              throw Error ("no digits after exponent sign in float literal")
          | _ => ()
        }
      };
      match (Char.ToLower (peek (), CultureInfo.InvariantCulture)) {
        | 'f' =>
          ignore (read ());
          Tok_float_literal (Single.Parse (buf.ToString (),
                                           NumberFormatInfo.InvariantInfo))
        | 'd' =>
          ignore (read ());          
          Tok_double_literal (Double.Parse (buf.ToString (),
                                           NumberFormatInfo.InvariantInfo))
        | 'm' =>
          ignore (read ());          
          Tok_decimal_literal (Decimal.Parse (buf.ToString (),
                                              NumberFormatInfo.InvariantInfo))
        | _ =>
          Tok_double_literal (Double.Parse (buf.ToString (),
                                            NumberFormatInfo.InvariantInfo))
      }
    };

    try {
      if (float_mode)
        match (peek ()) {
          | 'E' | 'e' => 
            exponent_part (false)
          | _ =>
            exponent_part (true)
        }
      else
        if (last_was_digit)
          match (Char.ToLower (peek (), CultureInfo.InvariantCulture)) {
            | 'e' =>
              exponent_part (false)
            | 'f' | 'd' | 'm' =>
              exponent_part (true)
            | _ =>
              /// we should have integer number here

              def special (c) {
                | 'l' | 's' | 'b' | 'u' =>
                  true
                | _ =>
                  false
              };

              // take read value as string
              def val = buf.ToString ();

              // check suffixes to make special types conversions
              mutable ch <- Char.ToLower (peek (), CultureInfo.InvariantCulture);
              if (special (ch)) {
                ignore (read ());
                mutable unsigned <- ch == 'u';

                // we can have two letter suffixes
                def ch' = Char.ToLower (peek (), CultureInfo.InvariantCulture);
                if (special (ch')) {
                  ignore (read ());
                  if (ch' == 'u') {
                    when (unsigned)
                      throw Error ("bad integer suffix (unsigned twice)");
                    unsigned <- true;
                  }
                  else {
                    // both chars can be different than 'u' and than they must be 'bs'
                    match ((ch, ch')) {
                      | ('s', 'b') | ('b', 's') =>
                        ch <- 'b'
                      | _ =>
                        unless (unsigned)
                          throw Error ("bad integer suffix (type contraint twice)");
                        ch <- ch';
                    }
                  }
                }
                else
                  // only 'b' suffix, so it is unsigned byte
                  when (ch == 'b') unsigned <- true;

                // [ch] now contains type meaning or 'u' (then it is uint for sure)
                match (ch) {
                  | 'b' =>
                    if (unsigned)
                      Tok_byte_literal (Convert.ToByte (val))
                    else
                      Tok_sbyte_literal (Convert.ToSByte (val))
                  | 's' =>
                    if (unsigned)
                      Tok_ushort_literal (Convert.ToUInt16 (val))
                    else
                      Tok_short_literal (Convert.ToInt16 (val))
                  | 'l' =>
                    if (unsigned)
                      Tok_ulong_literal (Convert.ToUInt64 (val))
                    else
                      Tok_long_literal (Convert.ToInt64 (val))
                  | _ =>
                    Tok_uinteger_literal (Convert.ToUInt32 (val))
                }
              }
              else
                Tok_integer_literal (Convert.ToInt32 (val))
          }
        else
          Tok_integer_literal (Convert.ToInt32 (buf.ToString ()))
    }
    catch {
      _ : System.OverflowException =>
        throw Error ("number literal " + buf.ToString () +
                     " is too large for given type")
    }
  }

  get_id (first_ch : char) : Token
  {
    if (first_ch == '\'' && !is_id_beg (peek ())) {
      get_char ()
    } else {
      def buf = Text.StringBuilder ();
      ignore (buf.Append (first_ch));

      while (is_id_beg (peek ()) || 
             Char.IsDigit (peek ()) || 
             peek () == '\'') {
        ignore (buf.Append (read ()))
      };
          
      def str = buf.ToString ();
      
      if (first_ch == '\'' && str.Length == 3 && str[2] == '\'')
        Tok_char_literal (str[1])
      else if (keywords.Contains (str))
        Tok_keyword (str)
      else
        Tok_identifier (str)
    }
  }

  escape_value (ch : char) : char
  {
    match (ch) {
      | 'n' => '\n'
      | 't' => '\t'
      | 'r' => '\r'
      | 'b' => '\b'
      | 'e' => '\e'
      | '"' => '"'
      | '\'' => '\''
      | '\\' => '\\'
      | '0' => '\0'
      | _ => throw Error ("bad escape character")
    }
  }
    
  get_char () : Token
  {
    match (get_string ('\'')) {
      | Tok_string_literal (v) =>
        if (v.Length == 0)
          throw Error ("empty character literal")
        else {
          if (v.Length != 1)
            throw Error ("character literal too long")
          else
            Tok_char_literal (v[0])
        }
      | _ => Util.ice ("Value is not string in get_char")
    }
  }

  get_string (end_ch : char) : Token
  {
    def buf = StringBuilder ();
    mutable go <- true;
              
    while (go) {
      match (read ()) {
        | '\\' =>
          def esc = escape_value (read ());
          ignore (buf.Append (esc));
        | '\n' =>
          throw Error ("newline before end of string")
        | ch when ch != end_ch => 
          ignore (buf.Append (ch))
        | _ =>
          def ch = eat_whitespace (' ');
          if (ch == end_ch) 
            when (comment_store.Length > 0) {
              Message.warning (get_location (),
                               "documentation comments between literals are ignored");
              ignore (comment_store.Remove (0, comment_store.Length))
            }
          else {
            isPendingChar <- true;
            pendingChar <- ch;
            go <- false
          }
      }
    };
     
    Tok_string_literal (buf.ToString ())
  }

  get_monkey_string () : Token
  {
    def buf = StringBuilder (32);

    def loop () {
      match (read ()) {
        | '"' => Tok_string_literal (buf.ToString ())
        | ch => ignore (buf.Append (ch)); loop ()
      }
    };

    loop ()
  }

  eat_whitespace (ch : char) : char
  {
    mutable eof <- false;
    
    def loop (c) {
      | ' ' | '\t' | '\r' | '\n' =>
        def ch = try read () catch { _ : Error => eof <- true; '_'  };
        if (eof) ' ' else loop (ch)
      | '/' =>
        match (peek ()) {
          | '/' =>
            // we are for sure in one line comment
            ignore (read ());
            if (peek () == '/') {
              ignore (read ());
              mutable cc <- ' ';
              do {
                cc <- read ();                
                ignore (comment_store.Append (cc))
              } while (cc != '\n')
            }
            else
              while (read () != '\n') {};
            // pass whitespace, so next read would be eof checked
            loop (' ')
              
          | '*' =>
            // multiline comment
            ignore (read ());
            def loop1 (seen_star, store) {
              def cc = read ();
              when (store)
                ignore (comment_store.Append (cc));
              
              match (cc) {
                | '*' => loop1 (true, store)
                | '/' when seen_star => ()
                | _ => loop1 (false, store)
              }
            };
            if (peek () == '*') {
              ignore (read ());
              loop1 (true, true);
              if (comment_store.Length == 1)
                ignore (comment_store.Remove (0, 1))
              else
                ignore (comment_store.Remove (comment_store.Length - 2, 2))
            }
            else
              loop1 (false, false);
            
            // pass whitespace, so next read would be eof checked            
            loop (' ')
            
          | _ => '/'
        }
      | _ => c
    };
    loop (ch)
  }
  
  do_get_token (ch : char) : Token
  {
    last_line <- line;
    last_col <- col;
    
    match (ch) {
      | '"' => get_string ('"')

      | '\'' => get_id ('\'')

      | '.' => 
        if (Char.IsDigit (peek ()))
          get_number (ch)
        else
          Tok_operator (ch.ToString (), 0)

      | '{' | '}' 
      | '[' | ']' 
      | ',' | ';' 
      | '(' | ')' => Tok_operator (String (ch, 1), 0)

      | '@' =>
        def next = read ();
        if (is_opchar (next))
          match (get_op (0, next)) {
            | Tok_operator (s, _) => Tok_identifier (s)
            | _ => Util.ice ()
          }
        else if (is_id_beg (next))
          match (get_id (next)) {
            | Tok_identifier (x)
            | Tok_keyword (x) => Tok_identifier (x)
            | _ => Util.ice ()
          }
        else if (next == '"')
          get_monkey_string ()
        else
          throw Error ("expecting operator, identifier or string literal after '@'")
      
      | '*' =>
        if (peek () == '*')
          get_op (1, ch)
        else
          get_op (2, ch)

      | '/' =>
        get_op (2, ch)
                              
      | '%' | '\\'
        => get_op (2, ch)

      | '-' =>
        if (Char.IsDigit (peek())) {
          negated_literal <- true;
          Tok_operator ("+", 3)
        }
        else
          if (peek () != '.') get_op (3, ch) 
          else Tok_operator (ch.ToString (), 3)          

/*
   THIS CODE TRIGGERS BUG IN MATCHING
      | '-' when Char.IsDigit (peek()) =>
        get_number (ch)
        
      | '+' | '-' 
        =>
           if (peek () != '.') get_op (3, ch) 
           else Tok_operator (ch.ToString (), 3)
*/             
      | '+'  
        =>
           if (peek () != '.') get_op (3, ch) 
           else Tok_operator (ch.ToString (), 3)
      
      | '^' | '$' | '~' | '?' | '#'
        => get_op (4, ch)
        
      | '=' | '<' | '>' | '!' 
        => get_op (5, ch)
        
      | '&' => get_op (6, ch)
      | '|' => get_op (7, ch)
      | ':' => get_op (8, ch)

      // after executing eat_whitespace it is the only possibility for space
      | ' ' => Tok_EOF ()
        
      | _ =>
        if (Char.IsDigit (ch))
          get_number (ch)
        else
          if (is_id_beg (ch))
            get_id (ch)
          else
            throw Error ("invalid character")
    }      
  }

  get_token () : Token implements ILexer.get_token
  {
    def ch = 
      if (isPendingChar) {
        isPendingChar <- false;
        eat_whitespace (pendingChar)
      }
      else
        eat_whitespace (' ');

    if (comment_store.Length > 0) {
      isPendingChar <- true;
      pendingChar <- ch;      
      
      def res = Tok_comment (comment_store.ToString ());
      ignore (comment_store.Remove (0, comment_store.Length));
      res
    }
    else
      do_get_token (ch)
  }

  get_location () : Location implements ILexer.get_location
  {
    Location (file_name, last_line, last_col)
  }

  get_current_location () : Location
  {
    Location (file_name, line, col)
  }

  public static this () 
  {
    def tab = array [
        "_", "abstract", "and", "array", "as", "base", "catch",
        "class", "def", "extends", "extern", "false", "finally", "fun",
        "implements", "interface", "internal", "macro", "match", "module",
        "mutable", "namespace", "new", "null", "out", "override",
        "params", "private", "protected", "public", "ref", "sealed",
        "static", "struct", "syntax", "this", "throw", "true", "try",
        "type", "typeof", "using", "variant", "virtual", "void", "when",
        "where"
    ];
    keywords <- Hashtable ();
    for (mutable i <- 0; i < tab.Length; ++i)
      keywords.Add (tab[i], 1)
  }

  public static AddKeyword (str : string) : void 
  {
    match (keywords.Get (str)) {
      | Some (level) =>
        keywords.Set (str, level + 1)
      | _ =>
        keywords.Set (str, 1)
    }
  }

  public static RemoveKeyword (str : string) : void 
  {
    match (keywords.Get (str)) {
      | Some (1) =>
        keywords.Remove (str)
      | Some (level) =>
        keywords.Set (str, level - 1)
      | _ =>
        Util.ice ("trying to remove non-existant keyword from lexer")
    }
  }
  
  public static IsOperator (str : string) : bool
  {
    def len = str.Length;
    mutable is <- true;
    for (mutable i <- 0; i < len && is; ++i) {
      def c = str[i];
      is <- is_opchar (c) || c == '(' || c == ')' || c == ';';
    };
    is
  }

  public static IsKeyword (str : string) : bool
  {
    def len = str.Length;
    mutable is <- true;
    for (mutable i <- 0; i < len && is; ++i)
      is <- Char.IsLetter (str[i]);
    is
  }
} // end class Lexer
} // end ns
